class RunDBInitException(Exception):
    pass


class RedundantForceException(Exception):
    pass


class NoNewerDumpAvailable(Exception):
    pass


class UnknownOpusAction(Exception):
    pass

class UnknownOpusUnit(Exception):
    pass

class EmploymentIdentifierNotUnique(Exception):
    pass


class ImporterrunNotCompleted(Exception):
    pass
